#!/bin/bash

print_help() {
    echo "Usage: $0 [-A] [-b] [-c <base_schema_file>] [-d] [-I <include_path>] [-p <platform>] [-v <vm_type>] <output_dir> <schema_file>"
    echo "  -A                      Allow the special character dot \".\" in an element key of MapV1/MapIterableV1."
    echo "  -b                      Serialize the schema file into a bianry buffer."
    echo "  -c <base_schema_file>   Check if the current schema is compatible to the base one."
    echo "  -d                      Use storage-through-dcvm mode in wasm contract. default: false"
    echo "  -I <include_path>       Set \"include path\" for \"include statements\" in the schema file."
    echo "  -p <platform>           Set target platform, available choices: [mychain, aldaba], default: mychain."
    echo "  -v <vm_type>            Set VM type, available choices: [wasm, dcvm], default: wasm."
    echo "  <output_dir>            Directory for output files."
    echo "  <schema_file>           Path to the input schema file."
}

print_error() {
    echo "\033[31m[ERROR]\033[0m $1"
    print_help
    exit 1
}

### Optional arguments
base_schema=""
allow_dot_in_map_key=""
binary_schema=""
storage_through_dcvm_mode=""
vm_type=""
include_paths=""
target_platform=""
while getopts ":Abc:dI:p:v:h" opt; do
    case $opt in
        A)
            allow_dot_in_map_key="--allow-dot-in-map-v1-key"
            ;;
        b)
            binary_schema="--binary --schema"
            ;;
        c)
            base_schema="--conform ${OPTARG}"
            ;;
        d)
            storage_through_dcvm_mode="--storage-through-dcvm"
            ;;
        I)
            include_paths+="-I ${OPTARG} "
            ;;
        p)
            target_platform="--platform ${OPTARG} "
            ;;
        v)
            vm_type="--vm-type ${OPTARG}"
            ;;
        h)
            print_help
            exit 0
            ;;
        *)
            print_error "Invalid option -${OPTARG}"
            ;;
    esac
done
shift $((OPTIND-1))

### Positional arguments
if [ $# -ne 2 ]; then
    print_error "There is $# positional arguments while expecting 2."
fi
out_dir="$1"
in_schema="$2"

### Commands
bindir=$(dirname ${BASH_SOURCE[0]})
include_paths+="-I ${bindir}/../wasm-sysroot/include/ "
include_configs="--keep-prefix"
${bindir}/myflatc --antcpp --gen-mutable --gen-object-api \
    $allow_dot_in_map_key $vm_type $storage_through_dcvm_mode \
    $target_platform \
    $include_configs $include_paths \
    -o "$out_dir" "$in_schema" $base_schema
if [ $? -ne 0 ]; then
    exit 1
fi
${bindir}/flatc-v1.12 --cpp $binary_schema --gen-mutable --gen-object-api \
    $include_configs $include_paths \
    -o "$out_dir" "$in_schema"
if [ $? -ne 0 ]; then
    exit 1
fi
