# 智能合约 ABI

## 版本变更记录

### abi/1.4  -->  abi/1.5
新增message字段用以存放开发者提供的合约字节码补充说明信息，长度为1024个字节

## 简介

ABI (Application Binary Interface) 描述了智能合约对外可见的接口，包括合约方法名和参数类型等。使用蚂蚁区块链的合约开发工具可以自动生成合约的 ABI 文件，以 Json 格式保存。

以下面的简单 C++ 合约为例：
```cpp
#include <mychainlib/contract.h>

using namespace mychain;

struct Name {
    std::string data;
    SERIALIZE(Name, (data));
};
typedef Name name_t;

CONTRACT_VERSION(123);

class Hello : public Contract {
public:
    INTERFACE int hi(name_t name) {
        print("Hello, %s", name.data.c_str());
        return 0;
    }
};

INTERFACE_EXPORT(Hello, (hi))
```

它对应的 ABI/1.5 文件如下：
```json
{
    "____comment": "This file was generated with wasm-abigen. DO NOT EDIT",
    "version": "mychain::abi/1.5",
    "contract_version": 123,
    "contract_git_id": "",
    "source_type": "C++",
    "vm_type": "wasm",
    "compiler_version": "my++ version nightly (77a5e1fc)",
    "structs": [
        {
            "name": "Name",
            "base": "",
            "fields": [
                {
                    "name": "data",
                    "type": "string"
                }
            ],
            "results": []
        },
        {
            "name": "hi",
            "base": "",
            "fields": [
                {
                    "name": "name",
                    "type": "name_t"
                }
            ],
            "results": ["int32"]
        }
    ],
    "message": "",
    "types": [
        {
            "new_type_name": "name_t",
            "type": "Name"
        }
    ],
    "interfaces": [
        {
            "name": "hi",
            "type": "hi"
        }
    ],
    "variants": []
}
```

它对应的 ABI/1.4 文件如下：
```json
{
    "____comment": "This file was generated with wasm-abigen. DO NOT EDIT",
    "version": "mychain::abi/1.4",
    "contract_version": 123,
    "contract_git_id": "",
    "source_type": "C++",
    "compiler_version": "my++ version 2.18-alpha (e684e181)",
    "structs": [
        {
            "name": "Name",
            "base": "",
            "fields": [
                {
                    "name": "data",
                    "type": "string"
                }
            ],
            "results": []
        },
        {
            "name": "hi",
            "base": "",
            "fields": [
                {
                    "name": "name",
                    "type": "name_t"
                }
            ],
            "results": ["int32"]
        }
    ],
    "types": [
        {
            "new_type_name": "name_t",
            "type": "Name"
        }
    ],
    "interfaces": [
        {
            "name": "hi",
            "type": "hi"
        }
    ],
    "variants": []
}
```

下文将分别介绍 ABI 中每个字段的含义。

## 合约元信息

### version
ABI 文件本身的版本。只有当 ABI 的格式有变化时（比如新增字段），这个版本号才会改变。

### contract_version
合约版本号。由合约代码中的 `CONTRACT_VERSION` 声明决定，默认值为 0。

### contract_git_id
如果合约代码文件处于一个 git 仓库中，这个字段会记录合约编译时该仓库的 git ID。

### source_type
合约源码类型，比如 C++ 或 Go。

### compiler_version
编译合约时所使用的编译器（如 my++）版本。

### message
abi/1.5版本新增
存放开发者提供的额外的合约字节码描述短信息，BMS（Blockchain Management System）系统在网页上展示字节码时，可以展示这部分信息作为合约补充说明信息。如果开发者没有提供该信息，则该字段为空。该字段信息唯一的输入方式为合约编译参数，示例命令为：`my++ hello.cpp -o hello.wasm -contract-message="This is a demo contract"`。请注意参数`-contract-message=`接受的信息长度为1024个字节。


## 合约接口描述

### Interfaces

每个 `interface` 对象描述了一个合约方法的名字和类型；合约方法类型包括了方法参数个数和各个参数的类型，具体定义可以在 `structs` 数组中找到（见下文）。例如上面的 ABI 中列出了合约方法 `hi`，其对应的类型描述也叫 `hi`。

### Structs

`struct` 对象可能表示两种类型：

1. 合约方法的类型

    `name` 为合约方法类型名（不要求和方法名相同），`fields` 为每个参数的名字和类型，`results` 表示合约方法返回值的类型。C++ 合约只有一个返回值，Go 合约可能有多个返回值。

2. 用户自定义结构体类型

    `name` 为结构体名字，`fields` 为结构体每个成员的名字和类型，`results` 字段为空。

如上面的例子中，`Name` 对象对应用户定义的结构体类型，`hi` 对象对应合约方法的接口类型。

### Types

当合约代码中使用 `typedef` 等语句定义了类型别名时，这些类型别名会记录到 `type` 对象中。

## 数据类型表示

ABI 中合约方法参数类型、结构体成员类型等的表示方式与合约代码中不一定完全相同。具体的对应关系如下：

### C++

| 源码 | ABI |
| --- | --- |
| `bool` | bool |
| `char, int8_t` | int8 |
| `unsigned char, uint8_t` | uint8 |
| `short, int16_t` | int16 |
| `unsigned short, uint16_t` | uint16 |
| `int, long, int32_t` | int32 |
| `unsigned int, unsigned long, uint32_t` | uint32 |
| `long long, int64_t` | int64 |
| `unsigned long long, uint64_t` | uint64 |
| `std::string` | string |
| `std::vector<std::string>` | string[] |
| `std::vector<AnyType>` | AnyType[] |

### Go

| 源码 | ABI |
| --- | --- |
| `bool` | bool |
| `int8` | int8 |
| `uint8, byte` | uint8 |
| `int16` | int16 |
| `uint16` | uint16 |
| `int32, int, rune` | int32 |
| `uint32, uint, uintptr` | uint32 |
| `int64` | int64 |
| `uint64` | uint64 |
| `string` | string |
| `[]string` | string[] |
| `[]AnyType` | AnyType[] |

注： `AnyType` 泛指任意类型
