#ifndef MYBUFFER_KEY_H
#define MYBUFFER_KEY_H

#include <memory>

#include <mybuffer/mychainlib.h>

constexpr char kDataContractMetaKeyPrefix[] = "root_data_contract";

inline void RevertOn(const std::shared_ptr<MychainLib>& _mychain_lib_ptr, bool condition, const std::string& message) {
    if (condition) {
        _mychain_lib_ptr->MYLOG(LogLevel::kError, message);
        _mychain_lib_ptr->MyRevert(message);
    }
}

enum struct FieldNodeType {TABLE, VECTOR, MAP_V1, MAP_V2};
inline std::string ToString(FieldNodeType node_type) {
  switch (node_type) {
    case FieldNodeType::TABLE : {
      return "TABLE";
    }
    case FieldNodeType::VECTOR : {
      return "VECTOR";
    }
    case FieldNodeType::MAP_V1 : {
      return "MAP_V1";
    }
    case FieldNodeType::MAP_V2 : {
      return "MAP_V2";
    }
    default : {
      return "Unknown Node Type: " + std::to_string(int(node_type));
    }
  }
}

class Varint {
public:
  static std::string Encode(uint64_t val) {
    val += 1; // escape '\0'
    if (val >= uint64_t('.')) {
      val += 1; // escape '.'
    }
    return EncodeInternal(val); // here, the range of val is [1, 45] + [47, UINT64_MAX]. (ascii('.') == 46)
  }

  static std::pair<uint64_t, size_t> Decode(const std::string& buf,
    const std::shared_ptr<MychainLib>& _mychain_lib_ptr) {
    auto result = DecodeInternal(buf, _mychain_lib_ptr);
    uint64_t val = result.first; // here, the expected range of val is [1, 45] + [47, UINT64_MAX]. (ascii('.') == 46)
    RevertOn(_mychain_lib_ptr, val == 0, "[ERROR] Decoded varint value is zero, from buf: " + buf);
    if (val-1 >= uint64_t('.')) {
      val -= 1; // de-escape '.'
    }
    val -= 1; // de-escape '\0'
    return {val, result.second};
  }

private:
  static std::string EncodeInternal(uint64_t val) {
    std::string buf;
    for (; val > 127; val >>= 7) {
      uint8_t b = 0x80 | uint8_t(val & 0x7F);
      buf.push_back(b);
    }
    buf.push_back(uint8_t(val));
    return buf;
  }

  static std::pair<uint64_t, size_t> DecodeInternal(const std::string& buf,
    const std::shared_ptr<MychainLib>& _mychain_lib_ptr) {
    uint64_t val = 0;
    size_t n=0;
    for (; n<buf.length() && (n*7 < 64); n++) {
      uint8_t b = buf[n];
      val |= uint64_t(b & 0x7F) << 7*n;
      if ((b & 0x80) == 0) {
       return {val, n+1};
      }
    }
    RevertOn(_mychain_lib_ptr, n==0, "[ERROR] Invalid string to decode as varint, empty");
    RevertOn(_mychain_lib_ptr, n!=0, "[ERROR] Invalid string to decode as varint, too long: " + buf);
    return {0, n}; // Never reach here
  }
};

class Key {
public:

  // For detailed information about key composition rule of v1 and v2,
  // check this table: https://yuque.antfin-inc.com/antchain/storage/dknyfl#EqxiD

  // For root Table node
  static std::unique_ptr<Key> get_root_table_key(const std::string& key,
    const std::shared_ptr<MychainLib>& _mychain_lib_ptr) {
    return std::unique_ptr<Key>(new Key(key, key+Varint::Encode(2), 1, _mychain_lib_ptr));
  }

  // For data contract meta node
  static std::unique_ptr<Key> get_data_contract_meta_key(const std::string& full_qualified_root_type_name,
    const std::shared_ptr<MychainLib>& _mychain_lib_ptr) {
    std::string key = ConcatWithUnderscore(kDataContractMetaKeyPrefix, full_qualified_root_type_name);
    return std::unique_ptr<Key>(new Key(key, key, 1, _mychain_lib_ptr));
  }

  // For non-root Table nodes, Vector nodes and  Map/MapIterable nodes
  std::unique_ptr<Key> get_table_field_key(FieldNodeType node_type, uint64_t field_id, const std::string& field_name="") const {
    int _key_version = 0;
    std::string _key_v1, _key_v2;

    switch (node_type) {
      case FieldNodeType::TABLE : {
        _key_v1 = "";
        _key_v2 = key_v2_ + Varint::Encode(field_id);
        _key_version = 1;
        break;
      }
      case FieldNodeType::VECTOR : {
        _key_v1 = _key_v2 = "";
        _key_version = 1;
        break;
      }
      case FieldNodeType::MAP_V1 : {
        _key_v1 = ConcatWithDot(key_v1_, field_name);
        _key_v2 = key_v2_ + Varint::Encode(field_id);
        _key_version = 1;
        break;
      }
      case FieldNodeType::MAP_V2 : {
        _key_v1 = ConcatWithDot(key_v1_, field_name);
        _key_v2 = key_v2_ + Varint::Encode(field_id);
        _key_version = 2;
        break;
      }
      default: {
        RevertOn(mychain_lib_ptr_, true, "Invalid node type for a field node in table: " + ToString(node_type));
      }
    }

    return std::unique_ptr<Key>(new Key(_key_v1, _key_v2, _key_version, mychain_lib_ptr_));
  }

  // For element nodes in a Vector
  std::unique_ptr<Key> get_vector_element_key() const {
    return std::unique_ptr<Key>(new Key("", "", 1, mychain_lib_ptr_));
  }

  // For element nodes in a Map/MapIterable
  std::unique_ptr<Key> get_map_element_key(const std::string& key) const {

    if (key_version_==1 && (!mychain_lib_ptr_->GetAllowDotInMapKey()) && (key.find(".") != std::string::npos)) {
      std::string msg = \
      "Dot \".\" is not allow in an element key ["+key+"] of MapV1/MapIterableV1"
        "\tTo skip this check, use `myflatc` executable with flag `--allow-dot-in-map-v1-key` or `myflatc.sh` script with option `-A` "
        "when compiling your schema file.";
      mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
      mychain_lib_ptr_->MyRevert(msg);
    }

    std::string _key_v1 = ConcatWithDot(key_v1_, key);
    std::string _key_v2 = key_v2_ + EncodeMapElementKey(key);
    return std::unique_ptr<Key>(new Key(_key_v1, _key_v2, key_version_, mychain_lib_ptr_));
  }

  // For index nodes in a MapIterable
  std::unique_ptr<Key> get_map_index_key(const std::string& key) const {
    // Index key composition of v2 with slash is safe,
    // always differs from Item key by starting with "/", rather than "L".
    std::string _key_v2 = ConcatWithSlash(key_v2_, key);
    std::string _key_v1 = ConcatWithSlash(key_v1_, key);
    return std::unique_ptr<Key>(new Key(_key_v1, _key_v2, key_version_, mychain_lib_ptr_));
  }

  // For meta_data node in a MapIterable
  std::unique_ptr<Key> get_map_meta_key() const {
    // An example of meta key of v1: root_contract_Transfer.my_map_iterable/M
    // An example of meta key of v2: root_contract_Transfer\x03\x01/M
    std::string key = "M";
    std::string _key_v2 = ConcatWithSlash(key_v2_, key);
    std::string _key_v1 = ConcatWithSlash(key_v1_, key);
    return std::unique_ptr<Key>(new Key(_key_v1, _key_v2, key_version_, mychain_lib_ptr_));
  }

  // For meta_data node in a Aldaba MapIterable
  std::unique_ptr<Key> get_aldaba_map_meta_key() const {
    // The key of meta_data node in an Aldaba MapIterable is the same as the MapIterable
    // An example of meta key of v1: root_contract_Transfer.my_map_iterable
    // An example of meta key of v2: root_contract_Transfer\x03\x01
    return std::unique_ptr<Key>(new Key(key_v1_, key_v2_, key_version_, mychain_lib_ptr_));
  }

  std::string to_string() const {
    if (key_version_ == 2) {
      RevertOn(mychain_lib_ptr_, key_v2_.empty(), "[ERROR] the key of version 2 is empty!");
      return key_v2_;
    } else if (key_version_ == 1) {
      return key_v1_;
    } else {
      RevertOn(mychain_lib_ptr_, true, "[ERROR] Not supported key version:" + std::to_string(key_version_));
      return "";
    }
  }

  const Key printable() const {
    return Key(key_v1_, to_printable(key_v2_), key_version_, mychain_lib_ptr_);
  }

  // Construct Key from existing key string.
  static std::unique_ptr<Key> from_existing_key_v2(const std::string& _key_v2, const std::shared_ptr<MychainLib>& _mychain_lib_ptr) {
    return std::unique_ptr<Key>(new Key("", _key_v2, 2, _mychain_lib_ptr));
  }

private:
  Key(const std::string& _key_v1, const std::string& _key_v2, int _key_version,
    const std::shared_ptr<MychainLib>& _mychain_lib_ptr)
    : key_v1_(_key_v1),
      key_v2_(_key_v2),
      key_version_(_key_version),
      mychain_lib_ptr_(_mychain_lib_ptr) {}

public:
  static std::string EncodeMapElementKey(const std::string& key) {
    return "L" + Varint::Encode(key.length()) + "C" + key;
  }

  static std::string ConcatWithDot(const std::string& first, const std::string& second) {
      return first + "." + second;
  }

  static std::string ConcatWithSlash(const std::string& first, const std::string& second) {
      return first + "/" + second;
  }

  static std::string ConcatWithUnderscore(const std::string& first, const std::string& second) {
      return first + "_" + second;
  }

  static std::string to_printable(
    const std::string& key_v2,
    std::shared_ptr<MychainLib> mychain_lib_ptr) {
    std::string buf;
    size_t step = 0;
    uint64_t val = 0;

    for (size_t i=0; i<key_v2.length(); i+=step) {
      // NOTE: if the varint encoded value is a printable char
      // it will not be decoded to its original value.
      if (std::isprint(int(key_v2[i])) == 0) {
        auto result = Varint::Decode(key_v2.substr(i), mychain_lib_ptr);
        val = result.first;
        step = result.second;
        buf += "\\" + std::to_string(val);
      } else {
        buf.push_back(key_v2[i]);
        step = 1;
      }
    }
    return buf;
  }

  std::string to_printable(const std::string& key_v2) const {
    return to_printable(key_v2, mychain_lib_ptr_);
  }

private:
  std::string key_v1_;
  std::string key_v2_;
  int key_version_;
  std::shared_ptr<MychainLib> mychain_lib_ptr_;
}; // class Key

#endif // MYBUFFER_KEY_H