#ifndef MYBUFFER_MYBASE_H
#define MYBUFFER_MYBASE_H

#include "mybuffer/key.h"
#include "mybuffer/mychainlib.h"

#include <map>
#include <memory>

class MyBase {
private:
    // It's kind of a singleton, which shared by all the serialization trees.
    class GlobalControlBlock {
    public:
        // Insert or update a root for later commitment.
        void upsert(const std::string &key, std::shared_ptr<MyBase> root) {
            roots_to_commit_[key] = root;
        }

        // Commit all the serialization trees
        int commit_all() {
            for (auto& key_and_root : roots_to_commit_) {
                int rv = key_and_root.second->commit_one();
                if (rv) return rv;
            }
            clear();
            return 0;
        }

        // Clean up all the serialization trees
        void clear() {
            roots_to_commit_.clear();
        }

        // For testing only
        size_t size() {
            return roots_to_commit_.size();
        }

    private:
        // For manual commit only, specified by MychainLib::GetManualCommit()
        // map<root_key, root_pointer>
        // Note: here we use shared pointers to keep the objects alive until they are committed.
        std::map<std::string, std::shared_ptr<MyBase> > roots_to_commit_;
    };

    // Each serialization tree shares a same control block.
    class ControlBlock {
    public:
        ControlBlock(const std::shared_ptr<MychainLib> &_mychain_lib_ptr)
            : mychain_lib_ptr_(_mychain_lib_ptr) {}

        ~ControlBlock() {}

        bool get_dirty() const {
            return dirty_;
        }

        void set_dirty(bool _dirty = true) {
            if (deleted_ && _dirty) {
                std::string msg = "Trying to modify deleted object.";
                mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
                mychain_lib_ptr_->MyRevert(msg);
                // if dirty is false, the modification is allowed.
            }
            dirty_ = _dirty;
        }

        bool get_deleted() const {
            return deleted_;
        }

        // Note: should only be called by a serialization tree root.
        void set_deleted(bool _deleted = true) {
            deleted_ = _deleted;
        }

        std::shared_ptr<MyBase> get_root_ptr() {
            std::shared_ptr<MyBase> rv = root_ptr_.lock();
            if (!rv) {
                std::string msg = "Failed to get the root pointer.";
                mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
                mychain_lib_ptr_->MyRevert(msg);
            }
            return rv;
        }

        void set_root_ptr(const std::shared_ptr<MyBase> &sp) {
            root_ptr_ = sp;
        }

    private:
        std::shared_ptr<MychainLib> mychain_lib_ptr_;
        bool dirty_ = false;
        bool deleted_ = false;
        // Point to the tree root without really owning it.
        std::weak_ptr<MyBase> root_ptr_;
    };

public:
    // All the constructors are protected and not allowed to be called outside.

    virtual ~MyBase() {}

    bool get_dirty() const {
        return control_block_ptr_->get_dirty();
    }

    // Note: this should be the only entrance to set "dirty_".
    void set_dirty(bool _dirty) {
        if (is_map_) {
            return;
        }
        control_block_ptr_->set_dirty(_dirty);
        if (_dirty && mychain_lib_ptr_->GetManualCommit()) {
            std::shared_ptr<MyBase> root_ptr = control_block_ptr_->get_root_ptr();
            global_control_block_ptr_->upsert(root_ptr->get_base_key(), root_ptr);
        }
    }

    bool get_deleted() const {
        return control_block_ptr_->get_deleted();
    }

    void set_deleted(bool value=true) {
        if (!is_root_) {
            std::string msg = "Failed to delete a non-root node: " + base_key_->to_string();
            mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
            mychain_lib_ptr_->MyRevert(msg);
        }
        control_block_ptr_->set_deleted(value);
    }

    // Note: it should be called after the following properties are set:
    // * is_root_
    void set_self_ptr(const std::shared_ptr<MyBase>& _self_ptr) {
        self_ptr_ = _self_ptr;
        if (is_root_) {
            control_block_ptr_->set_root_ptr(_self_ptr);
        }
    }

    std::shared_ptr<MyBase> get_self_ptr() {
        std::shared_ptr<MyBase> rv = self_ptr_.lock();
        if (!rv) {
            std::string msg = "Failed to get the self pointer.";
            mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
            mychain_lib_ptr_->MyRevert(msg);
        }
        return rv;
    }

    virtual int commit_one() {
        if (!is_root_) {
            std::string msg = "Failed to commit a non-root node: " + base_key_->to_string();
            mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
            mychain_lib_ptr_->MyRevert(msg);
        }
        if (!(mychain_lib_ptr_->GetManualCommit())) {
            std::string msg = "Tried to commit manually under auto-commit mode.";
            mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
            mychain_lib_ptr_->MyRevert(msg);
        }
        return 0;
    }

    int commit_all() {
        return global_control_block_ptr_->commit_all();
    }

    void clean_all() {
        global_control_block_ptr_->clear();
    }

    // For debugging, to print the path up to the root
    void print_path() const {
        mychain_lib_ptr_->MyPrint("PATH: %s", base_key_->to_string().data());
        if (parent_ptr_)
            parent_ptr_->print_path();
    }

    std::string get_base_key() const {
        return base_key_->to_string();
    }

    // For testing
    std::shared_ptr<MyBase> get_parent_ptr() const {
        return parent_ptr_;
    }

    // For testing
    size_t get_manual_commit_size() const {
        return global_control_block_ptr_->size();
    }

protected:
    MyBase(const std::shared_ptr<MychainLib>& _mychain_lib_ptr, std::unique_ptr<Key> _key, bool _is_root,
        const std::shared_ptr<MyBase>& _parent, bool _is_map=false)
        : mychain_lib_ptr_(_mychain_lib_ptr), base_key_(std::move(_key)), is_root_(_is_root), is_map_(_is_map) {
        if (is_root_) {
            control_block_ptr_ = new ControlBlock(_mychain_lib_ptr);
            control_block_mgr_.reset(control_block_ptr_);
            if (_parent) {
                parent_ptr_ = _parent;
                global_control_block_ptr_ = _parent->global_control_block_ptr_;
            } else {
                // Only the global root doesn't have parent.
                global_control_block_ptr_ = new GlobalControlBlock();
                global_control_block_mgr_.reset(global_control_block_ptr_);
            }
        } else {
            if (!_parent) {
                std::string msg = "Missing parent pointer for a non-root node.";
                mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
                mychain_lib_ptr_->MyRevert(msg);
            }
            parent_ptr_ = _parent;
            // ControlBlock is one-per-tree and GlobalControlBlock is one-for-all.
            if (!_is_map) {
                control_block_ptr_ = _parent->control_block_ptr_;
            }
            global_control_block_ptr_ = _parent->global_control_block_ptr_;
        }
    }

protected:
    std::shared_ptr<MychainLib> mychain_lib_ptr_;
    std::unique_ptr<Key> base_key_;
    bool is_root_ = false;
    bool is_map_ = false;

private:
    // Using weak_ptr downstream to traverse children
    std::weak_ptr<MyBase> self_ptr_;
    // Using shared_ptr upstream to make parents alive
    std::shared_ptr<MyBase> parent_ptr_;
    // For resource management of serialization tree roots only. Nullptr for others.
    std::unique_ptr<ControlBlock> control_block_mgr_;
    ControlBlock* control_block_ptr_;
    // For resource management of the global root only. Nullptr for others.
    std::unique_ptr<GlobalControlBlock> global_control_block_mgr_;
    GlobalControlBlock* global_control_block_ptr_;
};

#endif // MYBUFFER_MYBASE_H
