#ifndef MYBUFFER_MYMAP_H
#define MYBUFFER_MYMAP_H

#include "flatbuffers/flatbuffers.h"
#include "mybuffer/key.h"
#include "mybuffer/mybuffer_const_define.h"
#include "mybuffer/mybuffer_myfactory.h"
#include "mybuffer/mybuffer_mybase.h"
#include "mybuffer/mybuffer_mytable.h"
#include "mybuffer/mychainlib.h"

#include <map>
#include <memory>

template<class M>
class MyMap : public MyBase {
public:
    // No public constructors allowed. Use MyFactory to make instances instead.

    virtual ~MyMap() {}

    virtual bool has_element(std::string _key) {
        std::shared_ptr<M> sp = get_element(_key, false/*revert_on_failure*/);
        return (sp != nullptr) && !(sp->get_deleted());
    }

    // Two primary API's (get and add) to allocate memory for map elements
    // Don't call M::make_instance() directly elsewhere
    virtual std::shared_ptr<M> get_element(std::string _key, bool revert_on_failure=true, bool only_in_memory=false) {
        mychain_lib_ptr_->MYLOG(LogLevel::kDebug, "MyMap get_element: key %s", _key.c_str());
        std::shared_ptr<M> rv = get_element_in_cache(_key);
        // If the element is deleted, revert or return null;
        if (rv && rv->get_deleted()) {
            if (revert_on_failure) {
                std::unique_ptr<Key> key_obj = base_key_->get_map_element_key(_key);
                std::string msg = "Not allowed to get a deleted map element, key: "+ key_obj->printable().to_string();
                mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
                mychain_lib_ptr_->MyRevert(msg);
            } else {
                return nullptr;
            }
        }
        if (!rv && !only_in_memory) {
            std::unique_ptr<Key> key_obj = base_key_->get_map_element_key(_key);
            std::string full_key = key_obj->to_string();
            std::string buf;
            int ret = mychain_lib_ptr_->MyGetStorage(full_key, buf);
            if (ret != 0) {
                if (revert_on_failure) {
                    std::string msg = "Failed to get map element, key: " + key_obj->printable().to_string();
                    mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
                    mychain_lib_ptr_->MyRevert(msg);
                }
                else
                    return nullptr;
            } else {
                rv = MyFactory<M>::make_instance(mychain_lib_ptr_, buf, std::move(key_obj), get_self_ptr());
            }
            elements_[_key] = rv;
        }
        return rv;
    }

    virtual std::shared_ptr<M> add_element(std::string _key, bool do_check [[gnu::unused]] = true) {
        mychain_lib_ptr_->MYLOG(LogLevel::kInfo, "MyMap add_element: key %s", _key.c_str());
        // If the element already in cache, mark it as deleted, then erase it.
        std::shared_ptr<M> rv = get_element_in_cache(_key);
        if (rv) {
            // Mark the old element as "not dirty and deleted",
            // then it won't call MyDelStorage when in MaybePack.
            rv->set_dirty(false);
            rv->set_deleted(true);
            elements_.erase(_key);
        }

        std::unique_ptr<Key> key_obj = base_key_->get_map_element_key(_key);
        rv = MyFactory<M>::make_instance(mychain_lib_ptr_, std::move(key_obj), get_self_ptr());
        elements_[_key] = rv;
        return rv;
    }

    virtual int delete_element(std::string _key, bool revert_on_non_existence=true) {
        mychain_lib_ptr_->MYLOG(LogLevel::kInfo, "MyMap delete_element: key %s", _key.c_str());
        // No need to check if there are still any other shared_ptr owning the object here.
        // Instead, we will Revert() if later the user tries to modify the deleted object.
        std::shared_ptr<M> rv;
        if (revert_on_non_existence) {
            rv = get_element(_key, false/*revert_on_failure*/);
        } else {
            rv = get_element_in_cache(_key);
            if (!rv) {
                std::unique_ptr<Key> key_obj = base_key_->get_map_element_key(_key);
                rv = MyFactory<M>::make_instance(mychain_lib_ptr_, std::move(key_obj), get_self_ptr());
                elements_[_key] = rv;
            }
        }
        if (!rv || (revert_on_non_existence && rv && rv->get_deleted())) {
            std::unique_ptr<Key> key_obj = base_key_->get_map_element_key(_key);
            std::string msg = "Failed to delete map element, key: " + key_obj->printable().to_string();
            mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
            mychain_lib_ptr_->MyRevert(msg);
        }

        // If the element is already deleted, we do not need to change its status.
        if (!rv->get_deleted()) {
            rv->set_dirty(true); // set_dirty must be called on a non-deleted object(element).
            rv->set_deleted(true);
        }
        // Here, we do not erase rv from elements.
        return 0;
    }

protected:
    explicit MyMap(const std::shared_ptr<MychainLib>& _mychain_lib_ptr, std::unique_ptr<Key> _key,
        std::shared_ptr<MyBase> _parent)
        : MyBase(_mychain_lib_ptr, std::move(_key), false/*_is_root*/, _parent, true/*_is_map*/)  {
        static_assert(std::is_base_of<MyBase, M>::value, "Type M must derive from MyTable");
    }

    std::shared_ptr<M> get_element_in_cache(std::string _key) {
        std::shared_ptr<M> rv;
        auto itr = elements_.find(_key);
        if (itr != elements_.end()) {
            rv = itr->second.lock();
        }
        return rv;
    }

private:
    friend MyFactory<MyMap<M> >;

private:
    std::map<std::string, std::weak_ptr<M> > elements_;
};

#endif // MYBUFFER_MYMAP_H
