#ifndef MYBUFFER_MYTABLE_H
#define MYBUFFER_MYTABLE_H

#include "flatbuffers/flatbuffers.h"
#include "mybuffer/key.h"
#include "mybuffer/mybuffer_const_define.h"
#include "mybuffer/mybuffer_myfactory.h"
#include "mybuffer/mybuffer_mybase.h"
#include "mybuffer/mychainlib.h"

#include <memory>

template<class TNative, class TFlat, bool AsPassThroughBuffer = false, bool OnlyInMemory = false>
class MyTable : public MyBase {
public:
    // No public constructors allowed. Use MyFactory to make instances instead.

    virtual ~MyTable() {
        if (mychain_lib_ptr_->GetManualCommit()) {
            return;
        }
        MaybePack();
    }

    // Manual commit
    int commit_one() override {
        return MyBase::commit_one() || MaybePack();
    }

    // For testing only
    TNative* get_native_ptr() const {
        return native_table_ptr_;
    }

protected:
    // Constructing a serialization tree root from scratch.
    // It must be dirty since it is brand new.
    // _parent can only be nullptr for the global root.
    MyTable(const std::shared_ptr<MychainLib>& _mychain_lib_ptr, std::unique_ptr<Key> _key,
        const std::shared_ptr<MyBase>& _parent=nullptr)
        : MyBase(_mychain_lib_ptr, std::move(_key), true/*_is_root*/, _parent) {
        if (AsPassThroughBuffer) {
            pass_through_buffer_ = "";
        } else {
            native_table_mgr_.reset(new TNative);
            native_table_ptr_ = native_table_mgr_.get();
            Validate();
        }
    }

    // Constructing a serialization tree root with a flat buffer.
    // It starts as clean while initialized from disk.
    // _parent can only be nullptr for the global root.
    MyTable(const std::shared_ptr<MychainLib>& _mychain_lib_ptr, const std::string& _buf, std::unique_ptr<Key> _key,
        const std::shared_ptr<MyBase>& _parent=nullptr)
        : MyBase(_mychain_lib_ptr, std::move(_key), true/*_is_root*/, _parent) {
        if (AsPassThroughBuffer) {
            pass_through_buffer_ = _buf;
        } else {
            native_table_mgr_.reset((flatbuffers::GetRoot<TFlat>(_buf.data()))->UnPack());
            native_table_ptr_ = native_table_mgr_.get();
            Validate();
        }
    }

    // Constructing a non-root node with a native table.
    // The control block (control_block_) is usually from the parent MyTable.
    // The native table pointer (native_table_ptr_) is usually from the parent flatbuffers::NativeTable.
    explicit MyTable(const std::shared_ptr<MychainLib>& _mychain_lib_ptr, std::unique_ptr<Key> _key,
        std::unique_ptr<TNative>& _upnative, const std::shared_ptr<MyBase>& _parent)
        : MyBase(_mychain_lib_ptr, std::move(_key), false/*_is_root*/, _parent) {
        if (!_upnative) {
            _upnative.reset(new TNative);
        }
        native_table_ptr_ = _upnative.get();
        Validate();
    }

private:
    void Validate() {
        static_assert(std::is_base_of<flatbuffers::NativeTable, TNative>::value,
                      "Type TNative must derive from flatbuffers::NativeTable");
        static_assert(std::is_base_of<flatbuffers::Table, TFlat>::value,
                      "Type TFlat must derive from flatbuffers::Table");
        if (!native_table_ptr_) {
            std::string msg = "Missing the pointer to the native table.";
            mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
            mychain_lib_ptr_->MyRevert(msg);
        }
        if (is_root_ && !native_table_mgr_) {
            std::string msg = "Missing the native table manager for a serialization tree root";
            mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
            mychain_lib_ptr_->MyRevert(msg);
        }
    }

    int MaybePack() {
        if (OnlyInMemory) { return 0; }
        if (is_root_ && get_dirty()) {
            if (get_deleted()) {
                mychain_lib_ptr_->MYLOG(LogLevel::kInfo, "MyTable MaybePack: delete key %s", base_key_->to_string().c_str());
                mychain_lib_ptr_->MyDelStorage(base_key_->to_string());
            } else {
                if (AsPassThroughBuffer) {
                    mychain_lib_ptr_->MYLOG(LogLevel::kInfo, "MyTable MaybePack: pass-through set key %s", base_key_->to_string().c_str());
                    mychain_lib_ptr_->MySetStorage(base_key_->to_string(), pass_through_buffer_);
                } else {
                    mychain_lib_ptr_->MYLOG(LogLevel::kInfo, "MyTable MaybePack: non-pass-through set key %s", base_key_->to_string().c_str());
                    flatbuffers::FlatBufferBuilder fbb;
                    fbb.Finish(TFlat::Pack(fbb, native_table_ptr_));
                    std::string val((char *) fbb.GetBufferPointer(), fbb.GetSize());
                    mychain_lib_ptr_->MySetStorage(base_key_->to_string(), val);
                }
            }
        }
        return 0;
    }

    template <class M, class T>
    friend class MyFactory;

protected:
    TNative* native_table_ptr_;

    // pass-through buffer
    std::string pass_through_buffer_;

private:
    // For resource management of serialization tree roots only. Nullptr for others.
    std::unique_ptr<TNative> native_table_mgr_;
};

#endif // MYBUFFER_MYTABLE_H