#ifndef MYBUFFER_MYVECTOR_H
#define MYBUFFER_MYVECTOR_H

#include <memory>

#include "flatbuffers/flatbuffers.h"
#include "mybuffer/key.h"
#include "mybuffer/mybuffer_const_define.h"
#include "mybuffer/mybuffer_mybase.h"
#include "mybuffer/mybuffer_myfactory.h"
#include "mybuffer/mybuffer_mytable.h"
#include "mybuffer/mychainlib.h"

template <class M, class TNative, bool ForPrimitiveTypes = false>
class MyVector final : public MyBase {
 private:
  // When ForPrimitiveTypes is true, `M` equals `TNative`,
  // i.e. M == TNative == uint32_t
  //
  // Here, native_vector_type is just a reference type to flatbufers' vector.
  //
  // For vector of primitive types, we do not need a M-level `elements`
  // to keep weak-pointers to real objects.
  typedef
      typename std::conditional<ForPrimitiveTypes, std::vector<M>&,
                                std::vector<std::unique_ptr<TNative>>&>::type
          native_vector_type;

 public:
  // No public constructors allowed. Use MyFactory to make instances instead.

  ~MyVector() {}

  template <bool enabled = !ForPrimitiveTypes>
  typename std::enable_if<enabled, std::shared_ptr<M>>::type get_element(
      uint32_t idx) {
    mychain_lib_ptr_->MYLOG(LogLevel::kDebug, "MyVector get_element(non-primitive): index %d size %d", idx, native_elements_.size());
    if (idx >= native_elements_.size() || idx >= elements_.size()) {
      std::string msg = "Failed to get element because the index is out of range. given "
                        "index: " +
                        std::to_string(idx) + " size: " + std::to_string(elements_.size());
      mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
      mychain_lib_ptr_->MyRevert(msg);
    }

    std::shared_ptr<M> rv;
    if (!elements_[idx].expired()) {
      rv = elements_[idx].lock();
    }
    if (!rv) {
      rv = MyFactory<M, TNative>::make_instance(
          mychain_lib_ptr_, base_key_->get_vector_element_key(),
          native_elements_[idx], get_self_ptr());
      elements_[idx] = rv;
    }
    return rv;
  }

  template <bool enabled = ForPrimitiveTypes>
  typename std::enable_if<enabled, M>::type get_element(uint32_t idx) {
    mychain_lib_ptr_->MYLOG(LogLevel::kDebug, "MyVector get_element(primitive): index %d size %d", idx, native_elements_.size());
    if (idx >= native_elements_.size()) {
      std::string msg = "Failed to get element because the index is out of range. given "
                        "index: " +
                        std::to_string(idx) +
                        " size: " + std::to_string(native_elements_.size());
      mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
      mychain_lib_ptr_->MyRevert(msg);
    }

    return native_elements_[idx];
  }

  template <bool enabled = ForPrimitiveTypes>
  int set_element(uint32_t idx,
                  typename std::enable_if<enabled, M>::type value) {
    mychain_lib_ptr_->MYLOG(LogLevel::kDebug, "MyVector set_element(primitive): index %d size %d", idx, native_elements_.size());
    if (idx >= native_elements_.size()) {
      std::string msg = "Failed to set element because the index is out of range. given "
                        "index: " +
                        std::to_string(idx) +
                        " size: " + std::to_string(native_elements_.size());
      mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
      mychain_lib_ptr_->MyRevert(msg);
    }

    set_dirty(true);
    native_elements_[idx] = value;
    return 0;
  }

  uint32_t size() { return native_elements_.size(); }

  template <bool enabled = !ForPrimitiveTypes>
  typename std::enable_if<enabled, std::shared_ptr<M>>::type append_element() {
    uint32_t size = native_elements_.size();
    mychain_lib_ptr_->MYLOG(LogLevel::kDebug, "MyVector append_element(non-primitive): size %d", size);
    if (size >= size_limit_) {
      std::string msg = "Failed to append element because of vector size limit " +
                        std::to_string(size_limit_);
      mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
      mychain_lib_ptr_->MyRevert(msg);
    }

    native_elements_.emplace_back(new TNative);
    std::shared_ptr<M> rv = MyFactory<M, TNative>::make_instance(
        mychain_lib_ptr_, base_key_->get_vector_element_key(),
        native_elements_.back(), get_self_ptr());
    elements_.push_back(rv);
    return rv;
  }

  template <bool enabled = ForPrimitiveTypes>
  int append_element(typename std::enable_if<enabled, M>::type value) {
    uint32_t size = native_elements_.size();
    mychain_lib_ptr_->MYLOG(LogLevel::kDebug, "MyVector append_element(primitive): size %d", size);
    if (size >= size_limit_) {
      std::string msg = "Failed to append element because of vector size limit " +
                        std::to_string(size_limit_);
      mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
      mychain_lib_ptr_->MyRevert(msg);
    }

    set_dirty(true);
    native_elements_.emplace_back(value);
    return 0;
  }

  template <bool enabled = !ForPrimitiveTypes>
  typename std::enable_if<enabled, std::shared_ptr<M>>::type insert_element(
      uint32_t idx) {
    uint32_t size = native_elements_.size();
    mychain_lib_ptr_->MYLOG(LogLevel::kDebug, "MyVector insert_element(non-primitive): index %d size %d", idx, size);
    if (size >= size_limit_) {
      std::string msg = "Failed to insert element because of vector size limit " +
                        std::to_string(size_limit_);
      mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
      mychain_lib_ptr_->MyRevert(msg);
    }
    if (idx > size) {
      std::string msg = "Failed to insert element because the index is out of range.";
      mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
      mychain_lib_ptr_->MyRevert(msg);
    }
    native_elements_.emplace(native_elements_.begin() + idx, new TNative);
    std::shared_ptr<M> rv = MyFactory<M, TNative>::make_instance(
        mychain_lib_ptr_, base_key_->get_vector_element_key(),
        native_elements_[idx], get_self_ptr());
    elements_.insert(elements_.begin() + idx, rv);
    return rv;
  }

  template <bool enabled = ForPrimitiveTypes>
  int insert_element(uint32_t idx,
                     typename std::enable_if<enabled, M>::type value) {
    uint32_t size = native_elements_.size();
    mychain_lib_ptr_->MYLOG(LogLevel::kDebug, "MyVector insert_element(primitive): index %d size %d", idx, size);
    if (size >= size_limit_) {
      std::string msg = "Failed to insert element because of vector size limit " +
                        std::to_string(size_limit_);
      mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
      mychain_lib_ptr_->MyRevert(msg);
    }
    if (idx > size) {
      std::string msg = "Failed to insert element because the index is out of range.";
      mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
      mychain_lib_ptr_->MyRevert(msg);
    }
    set_dirty(true);
    native_elements_.emplace(native_elements_.begin() + idx, value);
    return 0;
  }

  int delete_element(uint32_t idx) {
    mychain_lib_ptr_->MYLOG(LogLevel::kDebug, "MyVector delete_element: index %d size %d", idx, native_elements_.size());
    if (idx >= native_elements_.size()) {
      std::string msg = "Failed to delete element because the index is out of range.";
      mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
      mychain_lib_ptr_->MyRevert(msg);
    }
    set_dirty(true);
    native_elements_.erase(native_elements_.begin() + idx);
    if (!ForPrimitiveTypes) {
      elements_.erase(elements_.begin() + idx);
    }
    return 0;
  }

  int clear() {
    mychain_lib_ptr_->MYLOG(LogLevel::kDebug, "MyVector clear: size %d", native_elements_.size());
    if (native_elements_.size() <= 0) {
      return 0;
    }
    set_dirty(true);
    native_elements_.clear();
    if (!ForPrimitiveTypes) {
      elements_.clear();
    }
    return 0;
  }

 protected:
  explicit MyVector(const std::shared_ptr<MychainLib>& _mychain_lib_ptr,
                    std::unique_ptr<Key> _key,
                    native_vector_type _native_elements,
                    std::shared_ptr<MyBase> _parent,
                    uint32_t _size_limit = kDefaultVectorSizeLimit)
      : MyBase(_mychain_lib_ptr, std::move(_key), false /*_is_root*/, _parent),
        size_limit_(_size_limit),
        native_elements_(_native_elements) {
    // We use "0" as an alias of the default value.
    if (size_limit_ == 0) {
      size_limit_ = kDefaultVectorSizeLimit;
    }
    // This can happen if a user has a big vector and then lowers the limit.
    if (native_elements_.size() > size_limit_) {
      std::string msg = "Failed to construct MyVector because of vector size limit " +
                        std::to_string(size_limit_);
      mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
      mychain_lib_ptr_->MyRevert(msg);
    }
    if (!ForPrimitiveTypes) {
      // Sync the size between native_elements_ and elements_
      elements_.resize(native_elements_.size());
    }
  }

 private:
  friend MyFactory<MyVector<M, TNative, ForPrimitiveTypes>, TNative>;

 private:
  uint32_t size_limit_;
  std::vector<std::weak_ptr<M>> elements_;
  // TODO: split the vector into a bunch of MyTable objects and get rid of
  // TNative here.
  native_vector_type native_elements_;
};

#endif  // MYBUFFER_MYVECTOR_H
