#ifndef MYBUFFER_MYCHAINLIB_H
#define MYBUFFER_MYCHAINLIB_H

#include <string>
#include <string.h>

#include "mybuffer/log.h"

// points to position right after the last occurrence of '/' in /__FILE__
#define __FILENAME__ strrchr("/" __FILE__, '/') + 1
#define MYLOG_HELPER(level, format, ...) MyPrintWithLevel(level, ("[%s:%d] " + std::string(format) + "%s").c_str(), __FILENAME__, __LINE__, __VA_ARGS__)
// add "" to handle variadic macros with zero arguments
#define MYLOG(level, ...) MYLOG_HELPER(level, __VA_ARGS__, "")

// Reference: https://yuque.antfin-inc.com/antchain/glrbr7/gm6ltv#bnvx2
enum MychainApiErrors {
    ERROR_NONE = 0,
    ERROR_DEPRECATING = 1,
    ERROR_NO_OBJECT = 1002,   // WS_NO_OBJECT_FAILED
    ERROR_KEY_DELETED = 1004,   // WS_OBJECT_DELETED_FAILED
    ERROR_KEY_NOT_EXISTS = 1011,   // WS_KEY_NOT_FOUND
    ERROR_DC_DATA_NOT_FOUND = 10760, // VM_DC_DATA_NOT_FOUND
    ERROR_OVER_SIZE = 10004,   // VM_MEMORY_ERROR
};

enum AldabaApiErrors {
    ALDABA_ERROR_NONE = 0,
    ALDABA_ERROR_KEY_NOT_EXISTS = 1,
};

class MychainLib {
public:
    virtual ~MychainLib() = default;
    virtual int MyPrint(const char* format, ...) = 0;
    virtual int MyPrintWithLevel(LogLevel level, const char* format, ...) = 0;
    virtual void MyRevert(const std::string& msg) = 0;

    bool GetAllowDotInMapKey() { return allow_dot_in_map_key_; }
    void SetAllowDotInMapKey(bool allow) { allow_dot_in_map_key_ = allow; }

    bool GetManualCommit() { return manual_commit_; }
    void SetManualCommit(bool val) { manual_commit_ = val; }

    LogLevel GetLogLevel() { return log_level_; }

    virtual uint32_t GetCurrentBlockVersion() { return 0; }
    virtual uint32_t GetBlockVersionOfReleaseTag(const std::string&) { return 1; }

    virtual int MyGetStorage(const std::string &key, std::string &val) {
        int rv = MyGetStorageInternal(key, val);
        if (rv != MychainApiErrors::ERROR_NONE) {
            if (rv != MychainApiErrors::ERROR_DEPRECATING &&
                rv != MychainApiErrors::ERROR_KEY_DELETED &&
                rv != MychainApiErrors::ERROR_KEY_NOT_EXISTS &&
                rv != MychainApiErrors::ERROR_DC_DATA_NOT_FOUND) {
                if (rv == MychainApiErrors::ERROR_OVER_SIZE) {
                    std::string msg = "Object key is over the 1KB limit.";
                    MYLOG(LogLevel::kError, msg);
                    MyRevert(msg);
                }
                else {
                    std::string msg = "Internal get-kv error: " + std::to_string(rv);
                    MYLOG(LogLevel::kError, msg);
                    MyRevert(msg);
                }
            }
            else {
                MYLOG(LogLevel::kWarning, "Failed to get storage for key %s error %d", key.c_str(), rv);
            }
        }
        return rv;
    }

    virtual int MySetStorage(const std::string &key, const std::string &val) {
        int rv = MySetStorageInternal(key, val);
        if (rv != MychainApiErrors::ERROR_NONE) {
            if (rv == MychainApiErrors::ERROR_OVER_SIZE) {
                std::string msg = "Object key/value is oversized.";
                MYLOG(LogLevel::kError, msg);
                MyRevert(msg);
            }
            else {
                std::string msg = "Internal set-kv error: " + std::to_string(rv);
                MYLOG(LogLevel::kError, msg);
                MyRevert(msg);
            }
        }
        return rv;
    }

    virtual int MyDelStorage(const std::string &key) {
        int rv = MyDelStorageInternal(key);
        if (rv != MychainApiErrors::ERROR_NONE) {
            if (rv == MychainApiErrors::ERROR_NO_OBJECT) {
                std::string msg = "Failed to delete a non-existing object.";
                MYLOG(LogLevel::kError, msg);
                MyRevert(msg);
            }
            else {
                std::string msg = "Internal del-kv error: " + std::to_string(rv);
                MYLOG(LogLevel::kError, msg);
                MyRevert(msg);
            }
        }
        return rv;
    }

protected:
    virtual int MyGetStorageInternal(const std::string &key, std::string &val) = 0;
    virtual int MySetStorageInternal(const std::string &key, const std::string &val) = 0;
    virtual int MyDelStorageInternal(const std::string &key) = 0;

private:
    bool allow_dot_in_map_key_ = false;
    // While manual commit enabled, we record all the changes and users need to call commit_all() to finalize.
    bool manual_commit_ = false;
    // Log_level's default value is set based off on this macro which can be defined via command line.
    LogLevel log_level_ = LogLevel(MYBUFFER_LOG_LEVEL);
};

#endif //MYBUFFER_MYCHAINLIB_H
