#ifndef MYFLATC_MYJSONGEN_H_
#define MYFLATC_MYJSONGEN_H_

#include "mybuffer/mychainlib.h"

#include <string>
#include <memory>

namespace mybuffer {

enum struct DecodeToJsonErrorCode {
  NONE,
  INVALID_BFBS_FILE_INVALID_FORMAT,
  INVALID_BFBS_FILE_NO_ROOT_TABLE,
  INVALID_BFBS_FILE_NOT_CONFORM,
  INVALID_KEY_FAILED_TO_DECODE,
  INVALID_KEY_V1_KEY,
  INVALID_KEY_MAP_ITERABLE_INDEX_KEY,
  INVALID_VALUE_FAILED_TO_DECODE,
};

enum struct DecoderVersion {
  Version2,  // Default version of the decoder
};

inline std::string ToString(DecoderVersion version) {
  switch (version) {
    case DecoderVersion::Version2:
      return "Version 2";
    default:
      return "Invalid Version: " + std::to_string(int(version));
  }
}

/**
 * @brief DecodeSchema() decodes a binary schema file to json format.
 * @param bfbsfile          The binary schema file content in string.
 * @param mychain_lib_ptr   The mychain library shared pointor.
 * @param version           The version number of current decoder.
 * @return                  Error code.
 * @return                  Decoded JSON text.
 */
extern std::pair<DecodeToJsonErrorCode, std::string> DecodeSchema(
    const std::string& bfbsfile, std::shared_ptr<MychainLib> mychain_lib_ptr,
    DecoderVersion version = DecoderVersion::Version2);

/**
 * @brief DecodeSchemaDelta() compares two binary schemas, then return the json
 * format of their delta.
 * @param bfbsfile_old      The base binary schema file content to be compared.
 * @param bfbsfile_new      The new binary schema file content.
 * @param mychain_lib_ptr   The mychain library shared pointor.
 * @param version           The version number of current decoder.
 * @return                  Error code.
 * @return                  Decoded JSON text.
 */
extern std::pair<DecodeToJsonErrorCode, std::string> DecodeSchemaDelta(
    const std::string& bfbsfile_old, const std::string& bfbsfile_new,
    std::shared_ptr<MychainLib> mychain_lib_ptr,
    DecoderVersion version = DecoderVersion::Version2);

/**
 * @brief DecodeValue() decodes the given mybuffer binary to JSON format by
 * parsing its binary schema file and mybuffer key.
 * @param bfbsfile          The binary schema file of the value.
 * @param key               The mybuffer key used to store the value.
 * @param value             The value.
 * @param mychain_lib_ptr   The mychain library shared pointor.
 * @param version           The version number of current decoder.
 * @return                  Error code.
 * @return                  Decoded JSON text.
 */
extern std::pair<DecodeToJsonErrorCode, std::string> DecodeValue(
    const std::string& bfbsfile, const std::string& key,
    const std::string& value, std::shared_ptr<MychainLib> mychain_lib_ptr,
    DecoderVersion version = DecoderVersion::Version2);

}  // namespace mybuffer

#endif  // MYFLATC_MYJSONGEN_H_
