/**
 * Copyright 2018 Ant Financial Group. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0(the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Author : dongwei.ldw@antfin.com
 * Date : 2018-12-28
 */

#pragma once
#include <string>
#include "datastream.hpp"
#include "developer_api.hpp"
#include "identity.hpp"
#include "types.h"
#include "evm.hpp"

#define INTERFACE [[mychain::interface]]
#define CONTRACT_VERSION(x) \
    [[maybe_unused]] const unsigned int __contract_version = (x)

#define CONTRACT_NAME(x) \
    [[maybe_unused]] const char __contract_name[] = x

namespace mychain {

#ifdef __cplusplus
extern "C" {
#endif

namespace detail {
int DelegateCall(const char* id_data,
                 uint32_t id_len,
                 const char* method_data,
                 uint32_t method_len,
                 const char* params_data,
                 uint32_t params_len);
int CallContract(const char* id_data,
                 uint32_t id_len,
                 const char* method_data,
                 uint32_t method_len,
                 uint64_t value,
                 uint64_t gas,
                 const char* params_data,
                 uint32_t params_len);
uint32_t GetCallResultSize();
int GetCallResult(char* out_data, uint32_t* out_len_ptr);
int __call_evm__(const char* id_ptr,
                 uint32_t id_len,
                 const char* method,
                 uint32_t method_len,
                 const unsigned char* input_ptr,
                 uint32_t input_len,
                 uint64_t gas,
                 uint64_t value);
}  // namespace detail

#ifdef __cplusplus
}
#endif

class Contract {
protected:
    virtual void MyChainSaveValue() {}
    virtual void MyChainLoadValue() {}

    template <typename T, typename... Args>
    inline decltype(auto) CallContract(const Identity& contract_id,
                                       const std::string& method,
                                       uint64_t value,
                                       uint64_t gas,
                                       Args... args) {
        MyChainSaveValue();

        auto params = pack(std::tuple<std::decay_t<Args>...>(args...));
        int code = detail::CallContract(
            contract_id.get_data().data(), contract_id.get_data().size(), method.data(),
            method.size(), value, gas, params.data(), params.size());

        // no matter call succeed or not, always get the result
        uint32_t val_len = detail::GetCallResultSize();
        std::string callResult(val_len, 0);
        int get_ret = detail::GetCallResult(callResult.data(), &val_len);
        if (get_ret) {
            code = 1;
            callResult = pack("get result failed"s);
        } else if (val_len > STORAGE_LIMIT) {
            code = 1;
            callResult = pack("bad result size"s);
        }
        auto ret = BuildCallResult<T>(code, callResult);

        MyChainLoadValue();
        return ret;
    }

    template <typename T, typename... Args>
    inline decltype(auto) DelegateCall(const Identity& contract_id,
                                       const std::string& method,
                                       Args... args) {
        MyChainSaveValue();
        auto params = pack(std::tuple<std::decay_t<Args>...>(args...));
        int code = detail::DelegateCall(contract_id.get_data().data(),
                                        contract_id.get_data().size(), method.data(),
                                        method.size(), params.data(), params.size());

        // no matter call succeed or not, always get the result
        uint32_t val_len = detail::GetCallResultSize();
        std::string callResult(val_len, 0);
        int get_ret = detail::GetCallResult(callResult.data(), &val_len);
        if (get_ret) {
            code = 1;
            callResult = pack("get result failed"s);
        } else if (val_len > STORAGE_LIMIT) {
            code = 1;
            callResult = pack("bad result size"s);
        }
        auto ret = BuildCallResult<T>(code, callResult);

        MyChainLoadValue();  // put it at end to avoid override call result.
        return ret;
    }

    static std::pair<int, bytes> CallEvm(Identity id,
                                         EVMParameter& p,
                                         uint64_t gas = 0,
                                         uint64_t value = 0) {
        int error_code = 1;
        bytes callResult;

        bytes input = p.GetData();
        std::string method = p.method_params_;
        error_code =
            detail::__call_evm__(id.get_data().data(), 32, method.c_str(),
                                 method.size(), input.data(), input.size(), gas, value);
        if (error_code != 0 && error_code != 10201) {
            return std::pair<int, bytes>(error_code, callResult);
        }
        uint32_t val_len = detail::GetCallResultSize();
        callResult.resize(val_len);
        int get_ret =
            detail::GetCallResult(reinterpret_cast<char*>(callResult.data()), &val_len);
        if (get_ret != 0) {
            Revert("GetCallResult fail, may be not enough memory.");
        }

        return std::pair<int, bytes>(error_code, callResult);
    }
};

}  // namespace mychain
