/**
 *  @file datastream.hpp
 *  @copyright defined in LICENSE.txt
 */
/**
 * Copyright 2018 Ant Financial Group. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0(the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Author : dongwei.ldw@antfin.com
 * Date : 2018-12-28
 */

#pragma once
#include <array>
#include <map>
#include <set>
#include <string>
#include <vector>

#include <stdio.h>
#include <boost/fusion/algorithm/iteration/for_each.hpp>
#include <boost/fusion/include/for_each.hpp>

#include <mychainlib/varint.hpp>

#ifdef datastream_test
#define stream_assert(condition, msg) \
    if (not(condition)) {             \
        printf("%s", (msg).c_str());  \
        assert(false);                \
    }
#else
#define stream_assert(condition, msg) MyAssert((condition), (msg));
#endif

namespace mychain {
extern void MyAssert(int, const std::string&);

/**
 *  %A data stream for reading and writing data in the form of std::string
 *  @brief %A data stream for reading and writing data in the form of std::string.
 *  @tparam T - Type of the datastream buffer
 */
template <typename D>
class datastream {
public:
    datastream(D start, size_t s, const std::string& err_prefix = "")
        : _start(start), _pos(start), _end(start + s), _err_prefix(err_prefix) {}

    inline void skip(size_t s) { _pos += s; }

    inline bool read(char* d, size_t s) {
        stream_assert(size_t(_end - _pos) >= (size_t)s,
                      _err_prefix + "failed to deserialize");
        memcpy(d, _pos, s);
        _pos += s;
        return true;
    }

    inline bool write(const char* d, size_t s) {
        stream_assert(_end - _pos >= (int32_t)s, _err_prefix + "failed to serialize");
        memcpy((void*)_pos, d, s);
        _pos += s;
        return true;
    }

    inline bool put(char c) {
        stream_assert(_pos < _end, _err_prefix + "failed to serialize");
        *_pos = c;
        ++_pos;
        return true;
    }

    inline bool unchecked_put(const char& c) {
        *_pos = c;
        ++_pos;
        return true;
    }

    inline bool get(unsigned char& c) { return get(*(char*)&c); }

    inline bool get(char& c) {
        stream_assert(_pos < _end, _err_prefix + "failed to deserialize");
        c = *_pos;
        ++_pos;
        return true;
    }

    D pos() const { return _pos; }
    inline bool valid() const { return _pos <= _end && _pos >= _start; }

    inline bool seekp(size_t p) {
        _pos = _start + p;
        return _pos <= _end;
    }

    inline size_t tellp() const { return size_t(_pos - _start); }

    inline size_t remaining() const { return _end - _pos; }

    inline const std::string& err_prefix() { return _err_prefix; }

private:
    D _start;
    D _pos;
    D _end;
    const std::string _err_prefix;
};

/**
 * @brief Specialization of datastream used to help determine the final size of
 * a serialized value. Specialization of datastream used to help determine the
 * final size of a serialized value
 */
template <>
class datastream<size_t> {
public:
    datastream(size_t init_size = 0) : _size(init_size) {}

    inline bool skip(size_t s) {
        _size += s;
        return true;
    }

    inline bool write(const char*, size_t s) {
        _size += s;
        return true;
    }

    inline bool put(char) {
        ++_size;
        return true;
    }

    inline bool valid() const { return true; }

    inline bool seekp(size_t p) {
        _size = p;
        return true;
    }

    inline size_t tellp() const { return _size; }

    inline size_t remaining() const { return 0; }

private:
    size_t _size;
};

namespace _datastream_detail {
// Check if type T is a pointer
template <typename T>
constexpr bool is_pointer() {
    return std::is_pointer<T>::value || std::is_null_pointer<T>::value ||
           std::is_member_pointer<T>::value;
}

// Check if type T is a primitive type
template <typename T>
constexpr bool is_primitive() {
    return std::is_arithmetic<T>::value || std::is_enum<T>::value;
}
}  // namespace _datastream_detail

// Pointer should not be serialized, so this function will always throws an error
template <typename D,
          typename T,
          std::enable_if_t<_datastream_detail::is_pointer<T>()>* = nullptr>
datastream<D>& operator<<(datastream<D>& ds, T) {
    static_assert(!_datastream_detail::is_pointer<T>(),
                  "Pointers should not be serialized");
    return ds;
}

// Pointer should not be deserialized, so this function will always throws an error
template <typename D,
          typename T,
          std::enable_if_t<_datastream_detail::is_pointer<T>()>* = nullptr>
datastream<D>& operator>>(datastream<D>& ds, T) {
    static_assert(!_datastream_detail::is_pointer<T>(),
                  "Pointers should not be deserialized");
    return ds;
}

//  Serialize a primitive type
template <typename D,
          typename T,
          std::enable_if_t<_datastream_detail::is_primitive<T>()>* = nullptr>
datastream<D>& operator<<(datastream<D>& ds, const T& v) {
    ds.write(reinterpret_cast<const char*>(&v), sizeof(T));
    return ds;
}

//  Deserialize a primitive type
template <typename D,
          std::enable_if_t<_datastream_detail::is_pointer<D>()>* = nullptr,
          typename T,
          std::enable_if_t<_datastream_detail::is_primitive<T>()>* = nullptr>
datastream<D>& operator>>(datastream<D>& ds, T& v) {
    ds.read(reinterpret_cast<char*>(&v), sizeof(T));
    return ds;
}

// Serialize a string into a stream
template <typename D>
datastream<D>& operator<<(datastream<D>& ds, const std::string& v) {
    ds << unsigned_int(v.size());
    if (v.size())
        ds.write(v.data(), v.size());
    return ds;
}

// Deserialize a string from a stream
template <typename D, std::enable_if_t<_datastream_detail::is_pointer<D>()>* = nullptr>
datastream<D>& operator>>(datastream<D>& ds, std::string& v) {
    unsigned_int s;
    ds >> s;
    v.resize(s);
    ds.read(v.data(), v.size());
    return ds;
}

// Serialize a fixed size array into a stream
template <typename D, typename T, std::size_t N>
datastream<D>& operator<<(datastream<D>& ds, const std::array<T, N>& v) {
    ds << unsigned_int(N);
    for (const auto& i : v)
        ds << i;
    return ds;
}

// Deserialize a fixed size array from a stream
template <typename D,
          std::enable_if_t<_datastream_detail::is_pointer<D>()>* = nullptr,
          typename T,
          std::size_t N>
datastream<D>& operator>>(datastream<D>& ds, std::array<T, N>& v) {
    unsigned_int s;
    ds >> s;
    stream_assert(N == s.value, ds.err_prefix() +
                                    "deserialize using datastream, std::array size and "
                                    "unpacked size don't match");
    for (auto& i : v)
        ds >> i;
    return ds;
}

// Serialize a vector of bool
template <typename D>
datastream<D>& operator<<(datastream<D>& ds, const std::vector<bool>& v) {
    ds << unsigned_int(v.size());
    bool b = false;
    for (size_t i = 0; i < v.size(); i++) {
        b = v[i];
        ds << b;
    }
    return ds;
}

//  Deserialize a vector of bool
template <typename D, std::enable_if_t<_datastream_detail::is_pointer<D>()>* = nullptr>
datastream<D>& operator>>(datastream<D>& ds, std::vector<bool>& v) {
    unsigned_int s;
    ds >> s;
    v.resize(s);
    bool b = false;
    for (size_t i = 0; i < v.size(); i++) {
        ds >> b;
        v[i] = b;
    }
    return ds;
}

// Serialize a vector of uint8_t
template <typename D>
datastream<D>& operator<<(datastream<D>& ds, const std::vector<uint8_t>& v) {
    ds << unsigned_int(v.size());
    ds.write(reinterpret_cast<const char*>(v.data()), v.size());
    return ds;
}

//  Deserialize a vector of uint8_t
template <typename D, std::enable_if_t<_datastream_detail::is_pointer<D>()>* = nullptr>
datastream<D>& operator>>(datastream<D>& ds, std::vector<uint8_t>& v) {
    unsigned_int s;
    ds >> s;
    v.resize(s);
    ds.read(reinterpret_cast<char*>(v.data()), v.size());
    return ds;
}

// Serialize a vector
template <typename D, typename T>
datastream<D>& operator<<(datastream<D>& ds, const std::vector<T>& v) {
    ds << unsigned_int(v.size());
    for (const auto& i : v)
        ds << i;
    return ds;
}

// Deserialize a vector
template <typename D,
          std::enable_if_t<_datastream_detail::is_pointer<D>()>* = nullptr,
          typename T>
datastream<D>& operator>>(datastream<D>& ds, std::vector<T>& v) {
    unsigned_int s;
    ds >> s;
    v.resize(s);
    for (auto& i : v)
        ds >> i;
    return ds;
}

//  Serialize a set
template <typename D, typename T>
datastream<D>& operator<<(datastream<D>& ds, const std::set<T>& s) {
    ds << unsigned_int(s.size());
    for (const auto& i : s) {
        ds << i;
    }
    return ds;
}

// Deserialize a set
template <typename D,
          std::enable_if_t<_datastream_detail::is_pointer<D>()>* = nullptr,
          typename T>
datastream<D>& operator>>(datastream<D>& ds, std::set<T>& s) {
    s.clear();
    unsigned_int sz;
    ds >> sz;

    for (uint32_t i = 0; i < sz; ++i) {
        T v;
        ds >> v;
        s.emplace(std::move(v));
    }
    return ds;
}

//  Serialize a map
template <typename D, typename K, typename V>
datastream<D>& operator<<(datastream<D>& ds, const std::map<K, V>& m) {
    ds << unsigned_int(m.size());
    for (const auto& i : m) {
        ds << i.first << i.second;
    }
    return ds;
}

//  Deserialize a map
template <typename D,
          std::enable_if_t<_datastream_detail::is_pointer<D>()>* = nullptr,
          typename K,
          typename V>
datastream<D>& operator>>(datastream<D>& ds, std::map<K, V>& m) {
    m.clear();
    unsigned_int s;
    ds >> s;

    for (uint32_t i = 0; i < s; ++i) {
        K k;
        V v;
        ds >> k >> v;
        m.emplace(std::move(k), std::move(v));
    }
    return ds;
}

//  Serialize a tuple
template <typename D, typename... Args>
datastream<D>& operator<<(datastream<D>& ds, const std::tuple<Args...>& t) {
    boost::fusion::for_each(t, [&](const auto& i) { ds << i; });
    return ds;
}

//  Deserialize a tuple
template <typename D,
          std::enable_if_t<_datastream_detail::is_pointer<D>()>* = nullptr,
          typename... Args>
datastream<D>& operator>>(datastream<D>& ds, std::tuple<Args...>& t) {
    boost::fusion::for_each(t, [&](auto& i) { ds >> i; });
    return ds;
}

template <typename D, typename T1, typename T2>
datastream<D>& operator<<(datastream<D>& ds, const std::pair<T1, T2>& p) {
    ds << p.first << p.second;
    return ds;
}

template <typename D, typename T1, typename T2>
datastream<D>& operator>>(datastream<D>& ds, std::pair<T1, T2>& p) {
    ds >> p.first >> p.second;
    return ds;
}

// Unpack data inside a fixed size buffer as T
template <typename T>
inline T unpack(const char* buffer, size_t len, const std::string& err_prefix = "") {
    T result;
    datastream<const char*> ds(buffer, len, err_prefix);
    ds >> result;
    return result;
}

template <typename T,
          typename P,
          std::enable_if_t<_datastream_detail::is_pointer<P>()>* = nullptr>
T unpack(P, const std::string&) {
    static_assert(!_datastream_detail::is_pointer<P>(), "Shouldn't unpack the pointer");
    return T();
}

// Unpack data inside a variable size buffer as T
template <typename T>
T unpack(const std::string& packed_data, const std::string& err_prefix = "") {
    return unpack<T>(packed_data.data(), packed_data.size(), err_prefix);
}

// Get the size of the packed data
template <typename T, std::enable_if_t<!_datastream_detail::is_pointer<T>()>* = nullptr>
size_t pack_size(const T& value) {
    datastream<size_t> ps;
    ps << value;
    return ps.tellp();
}

// Get packed data
template <typename T, std::enable_if_t<!_datastream_detail::is_pointer<T>()>* = nullptr>
std::string pack(const T& value, const std::string& err_prefix = "") {
    // FIXME(dongwei.ldw):
    // Root Cause:
    //      every elem in result are Assigned twice
    // ToDo:
    //      reduce twice to once
    //      marked on aone: https://aone.alipay.com/task/19159428
    std::string result(pack_size(value), 0);
    datastream<char*> ds(result.data(), result.size(), err_prefix);
    ds << value;
    return result;
}
}  // namespace mychain
