/**
 * Copyright 2018 Ant Financial Group. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0(the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Author : dongwei.ldw@antfin.com
 * Date : 2018-12-28
 */

#pragma once
#include <string>
#include "datastream.hpp"
#include "identity.hpp"
#include "tools.h"
#include "types.h"

namespace mychain {

#ifdef __cplusplus
extern "C" {
#endif

inline char* get_chain_data_buffer() {
    static char _chain_data_buffer[STORAGE_LIMIT];
    return _chain_data_buffer;
}

#pragma pack(push)
#pragma pack(1)
int print(const char* format, ...);
#pragma pack(pop)

namespace detail {
int Log(const char* user_custom_data,
        uint32_t user_custom_len,
        const char* topics_data,
        uint32_t topics_len);
}  // namespace detail

uint32_t FTraceBegin(const char* fname, uint32_t len);
uint32_t FTraceEnd(void);

#ifdef __cplusplus
}
#endif

bool CheckAccount(const Identity& id);

uint64_t GetBlockNumber();

int GetBlockHash(uint64_t block_number, std::string& block_hash);

uint64_t GetBlockTimeStamp();

Identity GetOrigin();

int GetAuthMap(const Identity& id, std::map<std::string, uint32_t>& auth);

int RetrieveSenderAuth(std::map<std::string, uint32_t>&);

int GetBalance(const Identity& id, uint64_t& value);

int GetCode(const Identity& id, std::string& code);

int GetCodeHash(const Identity& id, std::string& hash);

int GetRecoverKey(const Identity& id, std::string& recover_key);

int GetAccountStatus(const Identity& id, uint32_t& status);

std::string GetTxHash();

uint64_t GetGas();

bool IsLocalTx();

uint64_t GetValue();

std::string GetData();

int TransferBalance(const Identity& to, uint64_t balance);

int Require(bool condition, const std::string& str);

[[noreturn]] int Revert(const std::string& exception);

template <class T>
int Log(const T& data, const std::vector<std::string>& topics) {
    std::string buff = pack(data);

    // speed up packing topics
    size_t s = pack_size(unsigned_int(topics.size()));
    for (const auto& t : topics) {
        s += pack_size(unsigned_int(2 * t.size()));
        s += 2 * t.size();
    }
    std::string topics_b(s, 0);
    datastream<char*> ds(topics_b.data(),
                         topics_b.size());  // _err_prefix is not necessary
    ds << unsigned_int(topics.size());
    const char* dict_lower = "0123456789abcdef";
    for (const auto& t : topics) {
        ds << unsigned_int(2 * t.size());
        for (size_t i = 0; i < t.size(); i++) {
            char c = t[i];
            ds.unchecked_put(dict_lower[0x0f & (c >> 4)]);
            ds.unchecked_put(dict_lower[0x0f & c]);
        }
    }
    // Require(0 == ds.remaining(), "fail to pack topics");  // only used to debug

    int ret = detail::Log(buff.data(), buff.size(), topics_b.data(), topics_b.size());
    return ret ? 1 : 0;
}

template <typename T>
inline decltype(auto) BuildCallResult(int code, const std::string& callResult) {
    struct {
        int code;
        std::string msg;
        T result;
    } r;
    r.code = code;
    if (callResult.size() == 0) {
        // the result of a contract call might be empty
    } else if (code) {
        r.msg = unpack<std::string>(callResult, "MYCHAINLIB_ERROR: BuildCallResult ");
    } else {
        r.result = unpack<T>(callResult, "MYCHAINLIB_ERROR: BuildCallResult ");
    }
    return r;
}

template <>
inline decltype(auto) BuildCallResult<void>(int code, const std::string& callResult) {
    struct {
        int code;
        std::string msg;
    } r;
    r.code = code;
    if (callResult.size() == 0) {
        // the result of a contract call might be empty
    } else if (code) {
        r.msg = unpack<std::string>(callResult, "MYCHAINLIB_ERROR: BuildCallResult ");
    }
    return r;
}

Identity GetSelf();

Identity GetSender();

uint32_t DeployContract(const Identity& proposer_id,
                        const Identity& contract_id,
                        const std::string& contract_code,
                        const std::map<std::string, uint32_t>& authMap);

uint32_t UpdateContract(const Identity& contract_id,
                        const std::string& contract_code);

uint32_t UpdateContractStatus(const Identity& contract_id,
                              uint32_t status);

bool VerifyRsa(const std::string& pk,
               const std::string& sig,
               const std::string& msg,
               DigestType hash_type = SHA256);

bool VerifyMessageSM2(const std::string& pk,
                      const std::string& sig,
                      const std::string& msg);

bool Ecrecovery(const std::string& hash, const std::string& sig, Identity& id);

bool Digest(const std::string& msg, DigestType type, std::string& result);

bool Base64Encode(const std::string& input, std::string& output);

bool Base64Decode(const std::string& input, std::string& output);

// for pederson
#define RET_RANGEPROOF_SUCCESS 0X00000000
#define RET_RANGEPROOF_VERIFY_PASS 0X00000100
#define RET_RANGEPROOF_VERIFY_FAIL 0X00000101
#define RET_RANGEPROOF_ERROR 0XFFFFFFFF
#define RET_RANGEPROOF_ERROR_BAD_INPUT 0XFFFFFFFE
#ifndef ENABLE_GERMINI_UNITTEST
#define CONST_BLIND_FACTOR 3
#else
#define CONST_BLIND_FACTOR 0
#endif
typedef std::string PC;

/**
 * @brief range proof verify
 *
 * @param proof         [in    ] the proof value
 * @param pc_list       [in    ] pedersen commitment
 *
 * @return the status code.
 * @retval RET_RANGEPROOF_VERIFY_PASS
 * @retval RET_RANGEPROOF_VERIFY_FAIL
 * @retval RET_RANGEPROOF_ERROR_*
 */
int32_t RangeProofVerify(const std::string& proof, const std::vector<PC>& pc_list);

/**
 * @brief PC add
 *
 * @param PC_result     [   out] the result of PC add
 * @param PC_left       [in    ] left PC of Add
 * @param PC_right      [in    ] right PC of Add
 *
 * @return the status code.
 * @retval RET_RANGEPROOF_SUCCESS success
 * @retval RET_RANGEPROOF_ERROR_*
 *
 */
int32_t AddPedersenCommit(PC& PC_result, const PC& PC_left, const PC& PC_right);

/**
 * @brief PC sub
 *
 * @param PC_result     [   out] the result of PC sub
 * @param PC_left       [in    ] left PC of sub
 * @param PC_right      [in    ] right PC of sub
 *
 * @return the status code.
 * @retval RET_RANGEPROOF_SUCCESS success
 * @retval RET_RANGEPROOF_ERROR_*
 *
 */
int32_t SubPedersenCommit(PC& PC_result, const PC& PC_left, const PC& PC_right);

/**
 * @brief PC calculate
 *
 * @param dst_pc      [   out] the PC result
 * @param src_pc      [in    ] the initialize value of PC
 * @param positive    [in    ] the positive PC list
 * @param negative    [in    ] the negative PC list
 *
 * @return the status code.
 * @retval RET_RANGEPROOF_SUCCESS success
 * @retval RET_RANGEPROOF_ERROR_*
 */
int32_t CalculatePedersenCommit(PC& dst_pc,
                                const PC& src_pc,
                                const std::vector<PC>& positive,
                                const std::vector<PC>& negative);

/**
 * @brief PC equality verify
 *
 * @param positive    [in    ] the positive PC list
 * @param negative    [in    ] the negative PC list
 *
 * @return the status code.
 * @retval RET_RANGEPROOF_VERIFY_PASS
 * @retval RET_RANGEPROOF_VERIFY_FAIL
 * @retval RET_RANGEPROOF_ERROR_*
 */
int32_t PedersenCommitEqualityVerify(const std::vector<PC>& positive,
                                     const std::vector<PC>& negative);

bool VerifyMessageECCK1(const std::string& pubkey,
                        const std::string& sig,
                        const std::string& msg,
                        DigestType hash_type);

bool VerifyMessageECCR1(const std::string& pubkey,
                        const std::string& sig,
                        const std::string& msg,
                        DigestType hash_type);

int CreateContract(const Identity& src, const Identity& new_contract);

uint32_t GetRelatedTransactionListSize(const Identity& recipient_id,
                                       uint64_t deposit_flag,
                                       uint64_t& count);

uint32_t GetRelatedTransactionList(const Identity& recipient_id,
                                   uint64_t deposit_flag,
                                   uint64_t start_index,
                                   uint32_t number,
                                   std::vector<std::string>& tx_list);

uint32_t GetTransactionSender(const std::string& tx_hash, Identity& sender);
uint32_t GetTransactionReceiver(const std::string& tx_hash, Identity& receiver);
uint32_t GetTransactionTimestamp(const std::string& tx_hash, uint64_t& timestamp);
uint32_t GetTransactionData(const std::string& tx_hash, std::string& data);
uint32_t GetTransactionBlockIndex(const std::string& tx_hash,
                                  uint64_t& block_number,
                                  uint32_t& index);
uint32_t GetTransactionDepositFlag(const std::string& tx_hash, uint64_t& deposit_flag);

uint32_t GetConfidentialDepositData(const std::string& tx_hash, std::string& data);
uint32_t GetConfidentialDepositFlag(const std::string& tx_hash, uint64_t& deposit_flag);

/**
 * Homomorphic add for lifted elgamal encryption, res = first + second
 *
 * @param first     [IN]       lifted elgamal ciphertext，68 byte
 * @param second    [IN]      lifted elgamal ciphertext， 68 byte
 * @param res       [OUT]       lifted elgamal ciphertext, used to store result，68 byte
 * @return      1 on success,
                0 on error,
                2000 on NULL parameter
 */
int LiftedElgamalContractHomomorphicAdd(const std::string& first,
                                        const std::string& second,
                                        std::string& res);

/**
 * Homomorphic sub for lifted elgamal encryption，res = first - second
 *
 * @param first       [IN]      lifted elgamal ciphertext, 68 byte
 * @param second      [IN]      lifted elgamal ciphertext， 68 byte
 * @param res         [OUT]     lifted elgamal ciphertext, used to store result，68 byte
 * @return      1 on success,
                0 on error,
                2000 on NULL parameter
 */
int LiftedElgamalContractHomomorphicSub(const std::string& first,
                                        const std::string& second,
                                        std::string& res);

/**
 * Homomorphic mutiply for lifted elgamal encryption，res = first * second
 *
 * @param src         [IN]      lifted elgamal ciphertext, 68 byte
 * @param scalar      [IN]      multiplier, uint64_t scalar
 * @param res         [OUT]     lifted elgamal ciphertext, used to store result，68 byte
 * @return      1 on success,
                0 on error,
                2000 on NULL parameter
 */
int LiftedElgamalScalarMutiply(const std::string& src,
                               uint64_t scalar,
                               std::string& res);

int LiftedElgamalScalarMultiply(const std::string& src,
                                uint64_t scalar,
                                std::string& res);

/**
 * Verify zero check proof
 *
 * @param ciphertext     [IN]      Elgamal ciphertext, 68 byte
 * @param proof          [IN]      Zero check proof
 * @param verify_result  [OUT]     verify result, 1 on success, 0 on failed
 * @return                  1 on success,
                            0 on error,
                            10001 on NULL parameter
 */

int LiftedElgamalContractZeroCheckVerify(const std::string& ciphertext,
                                         const std::string& proof,
                                         int& verify_result);

/**
 * Verify range proof
 * @param ciphertext           [IN]   Elgamal ciphertext list (array)，each 68 byte
 * @param per_value_bit_size   [IN]   Length of plaintext space, described in bits
 * @param public_key           [IN]   Public key， 33 byte
 * @param proof                [IN]   Range proof
 * @param verify_result        [OUT]  verify result, 1 on success, 0 on failed
 * @return                     1 on success,
                               0 on error,
                               10001 on NULL parameter
 */
int LiftedElgamalContractRangeVerify(const std::vector<std::string>& ciphertext,
                                     const int per_value_bit_size,
                                     const std::string& public_key,
                                     const std::string& proof,
                                     int& verify_result);

/**
 * Verify snark proof
 * @param verification_key     [IN]   Zk-snark verification key
 * @param inputs               [IN]   Zk-snark public inputs
 * @param proof                [IN]   Zk-snark proof
 * @return                     0 on success,
                               1 on error,
                               10001 on NULL parameter
 */

int BellmanSnarkVerify(const std::string& verification_key,
                       const std::vector<std::string>& inputs,
                       const std::string& proof);

uint32_t DCGetStorage(const Identity& dc_id,
                      uint32_t db_id,
                      const std::string& key,
                      std::string& val);

uint32_t DCSetStorage(const Identity& dc_id,
                      uint32_t db_id,
                      const std::string& key,
                      const std::string& val);

uint32_t DCSetAcl(const Identity& contract_id,
                  const std::vector<std::pair<Identity, uint8_t>>& objects,
                  uint32_t db_id = 0);

uint32_t DCDeleteStorage(const Identity& dc_id, uint32_t db_id, const std::string& key);

uint32_t GrayscaleDeployContract(const Identity& contract_id,
                                 const std::string& contract_code);

uint32_t GrayscaleVerification(const Identity& contract_id);

uint32_t GrayscaleVersionSwitchBack(const Identity& contract_id);

uint32_t GrayscaleUpdateContract(const Identity& contract_id);

bool IsAccountExist(const Identity& id);

/**
 *
 *  * @param in targetName: the target name to be digest
 *
 *   * @param out id: the platform specific digest result on success, it should be unchanged on failure.
 *
 *    * @return true on success, false on failure.
 *
 *     */

bool GetTargetAddress(std::string targetName, Identity& id);

}  // namespace mychain
