/**********************************************************************
 * Copyright (C) 2020 Ant Financial. All rights reserved.
 *
 * FileName : parameter_evm.cpp
 * Desc     :
 * History  :
 * License  : Ant Financial proprietary/confidential.
 **********************************************************************/
#ifndef PARAMETER_EVM_HPP
#define PARAMETER_EVM_HPP
#include <map>
#include <string>
#include <vector>
#include <byteswap.h>
#include "vm_util.hpp"

namespace mychain {

using byte = uint8_t;
using bytes = std::vector<byte>;
using uint = unsigned int;

static std::string Bytes2Hex(const bytes& input, bool uppercase = false) {
    const char* dict_lower = "0123456789abcdef";
    const char* dict_upper = "0123456789ABCDEF";
    const char* dict = uppercase ? dict_upper : dict_lower;
    std::string out;
    size_t input_size = input.size();
    out.reserve(input_size * 2);
    int h = 0, l = 0;
    for (size_t i = 0; i < input_size; i++) {
        int c = input[i];
        h = 0x0f & (c >> 4);
        out.push_back(dict[h]);
        l = 0x0f & c;
        out.push_back(dict[l]);
    }
    return out;
}

static bytes str2Bytes(const std::string& x) {
    bytes ret;
    ret.resize(x.size());
    memcpy(ret.data(), reinterpret_cast<const uint8_t*>(x.data()), x.size());
    return ret;
}

static bytes int2Bytes(uint64_t x) {
    bytes ret;
    ret.resize(32);
    uint64_t y = bswap_64(x);
    memcpy(ret.data() + 24, &y, 8);
    return ret;
}

class Parameter {
public:
    Parameter() = default;
    virtual ~Parameter() = default;

public:
    virtual bytes GetData() = 0;
};

class VMOutput {
public:
    VMOutput() = default;
    virtual ~VMOutput() = default;

public:
    inline void SetOutput(const bytes& output) { output_ = output; }
    inline void SetOffset(uint offset) { offset_ = offset; }

    inline bytes& GetOutput() { return output_; }
    inline uint& GetOffset() { return offset_; }

protected:
    bytes output_;    // Output of the code
    uint offset_{0};  // Parse the offset of output
};

class EVMParameter : public Parameter {
public:
    EVMParameter() = default;
    ~EVMParameter() {}

public:
    void SetFunctionSelector(const std::string& method_params) {
        if (!method_params.empty()) {
            method_params_ = method_params;
        }
    }

    bytes GetData() override {
        bytes data;
        // std::string hash;
        // Digest(method_params_, DigestType::KECCAK, hash);
        // data.resize(4);
        // memcpy(data.data(), reinterpret_cast<uint8_t*>(hash.data()), 4);

        int offset = 32 * (static_params_.size() + dynamic_params_.size());
        for (int i = 0; i < param_num_; ++i) {
            if (static_params_.find(i) != static_params_.end()) {
                data.insert(data.end(), static_params_[i].begin(),
                            static_params_[i].end());
            } else {
                bytes offset_encode = int2Bytes(offset);
                data.insert(data.end(), offset_encode.begin(), offset_encode.end());
                offset += dynamic_params_[i].size();
            }
        }

        for (std::map<int, bytes>::iterator iter = dynamic_params_.begin();
             iter != dynamic_params_.end(); ++iter) {
            data.insert(data.end(), iter->second.begin(), iter->second.end());
        }

        return data;
    }

    void AddInt(int64_t value) {
        bytes tmp = int2Bytes(value);
        static_params_[param_num_++] = tmp;
    }

    void AddUInt(uint64_t value) {
        bytes tmp = int2Bytes(value);
        static_params_[param_num_++] = tmp;
    }

    void AddBool(bool value) {
        bytes tmp;
        tmp.resize(32);
        tmp[31] = value;
        static_params_[param_num_++] = tmp;
    }

    void AddIdentity(const Identity& value) {
        bytes tmp;
        std::string d = value.get_data();
        tmp.assign(d.data(), d.size() + d.data());
        static_params_[param_num_++] = tmp;
    }

    void AddString(const std::string& value) {
        bytes data = int2Bytes(value.size());
        data.resize(data.size() + (value.size() + 31) / 32 * 32);
        // bytesConstRef(&value).populate(bytesRef(&data).cropped(32)); TODO
        memcpy(data.data() + 32, value.data(), value.size());
        dynamic_params_[param_num_++] = data;
    }

    void AddBytes(const bytes& value) {
        bytes tmp = int2Bytes(value.size());
        tmp.resize(tmp.size() + (value.size() + 31) / 32 * 32);
        memcpy(tmp.data() + 32, value.data(), value.size());
        dynamic_params_[param_num_++] = tmp;
    }

    void AddBytesN(const bytes& value) {
        bytes tmp = value;
        tmp.resize(32);
        static_params_[param_num_++] = tmp;
    }

    void AddIntArray(const std::vector<int64_t>& value) {
        bytes data;
        for (int i = 0; i < static_cast<int>(value.size()); i++) {
            bytes b = int2Bytes(value[i]);
            data.insert(data.end(), b.begin(), b.end());
        }

        bytes tmp = int2Bytes(value.size());
        tmp.insert(tmp.end(), data.begin(), data.end());
        dynamic_params_[param_num_++] = tmp;
    }

    void AddBytesNArray(const std::vector<bytes>& value) {
        bytes data;
        for (size_t i = 0; i < value.size(); i++) {
            bytes tmp = value[i];
            tmp.resize(32);
            data.insert(data.end(), tmp.begin(), tmp.end());
        }
        dynamic_params_[param_num_++] = data;
    }

    void AddUIntArray(const std::vector<uint64_t>& value) {
        bytes data;
        for (int i = 0; i < static_cast<int>(value.size()); i++) {
            bytes r = int2Bytes(value[i]);
            data.insert(data.end(), r.begin(), r.end());
        }

        bytes tmp = int2Bytes(value.size());

        tmp.insert(tmp.end(), data.begin(), data.end());
        dynamic_params_[param_num_++] = tmp;
    }

    void AddBoolArray(const std::vector<bool>& value) {
        bytes data;
        for (int i = 0; i < static_cast<int>(value.size()); i++) {
            auto r = int2Bytes(uint64_t(value[i]));
            data.insert(data.end(), r.begin(), r.end());
        }

        bytes tmp = int2Bytes(value.size());
        tmp.insert(tmp.end(), data.begin(), data.end());
        dynamic_params_[param_num_++] = tmp;
    }

    void AddStringArray(const std::vector<std::string>& value) {
        bytes out;
        size_t value_count = value.size();
        uint64_t total_solt = 1 + value_count;  // count for value.size

        out.resize(32 * total_solt);

        bytes tmp = int2Bytes(value_count);
        memcpy(out.data(), tmp.data(), 32);

        for (int i = 0; i < value_count; i++) {
            const std::string& item = value[i];
            uint64_t item_slot = ((item.size() + 31) / 32);
            total_solt += (1 + item_slot);
        }

        out.resize(total_solt * 32);
        uint64_t offset = 32 * (value_count);
        for (uint i = 0; i < value_count; i++) {
            const std::string& item = value[i];
            uint64_t item_slot = ((item.size() + 31) / 32);
            memcpy(out.data() + 32 * (i + 1), int2Bytes(offset).data(), 32);
            memcpy(out.data() + offset + 32, int2Bytes(item.size()).data(), 32);
            memcpy(out.data() + offset + 64, item.data(), item.size());
            offset += 32;
            offset += 32 * (item_slot);
        }

        dynamic_params_[param_num_++] = out;
    }
    bool has_fail_ = false;

    std::string method_params_;

private:
    int param_num_ = 0;
    std::map<int, bytes> static_params_;
    std::map<int, bytes> dynamic_params_;
};

class EVMOutput : public VMOutput {
public:
    EVMOutput() = default;
    ~EVMOutput() {}

public:
    bool ReadInt(int64_t& value) {
        if (VMUtil::ReturnIntValue(output_, offset_, value)) {
            offset_++;
            return true;
        }

        return false;
    }

    bool ReadUInt(uint64_t& value) {
        if (VMUtil::ReturnUintValue(output_, offset_, value)) {
            offset_++;
            return true;
        }

        return false;
    }

    bool ReadBool(bool& value) {
        if (VMUtil::ReturnBoolValue(output_, offset_, value)) {
            offset_++;
            return true;
        }

        return false;
    }

    bool ReadIdentity(Identity& value) {
        if (VMUtil::ReturnIdentityValue(output_, offset_, value)) {
            offset_++;
            return true;
        }

        return false;
    }

    bool ReadString(std::string& value) {
        if (VMUtil::ReturnStringValue(output_, offset_, value)) {
            offset_++;
            return true;
        }

        return false;
    }

    bool ReadBytes(bytes& value) {
        if (VMUtil::ReturnBytesValue(output_, offset_, value)) {
            offset_++;
            return true;
        }

        return false;
    }

    bool ReadBytesN(bytes& value) {
        if (VMUtil::ReturnBytesNValue(output_, offset_, value)) {
            offset_++;
            return true;
        }

        return false;
    }

    bool ReadIntArray(std::vector<int64_t>& value) {
        if (VMUtil::ReturnIntArrayValue(output_, offset_, value)) {
            offset_++;
            return true;
        }

        return false;
    }

    bool ReadUIntArray(std::vector<uint64_t>& value) {
        if (VMUtil::ReturnUintArrayValue(output_, offset_, value)) {
            offset_++;
            return true;
        }

        return false;
    }

    bool ReadBytesNArray(std::vector<bytes>& value) {
        if (VMUtil::ReturnBytesNArrayValue(output_, offset_, value)) {
            offset_++;
            return true;
        }

        return false;
    }

    bool ReadBoolArray(std::vector<bool>& value) {
        if (VMUtil::ReturnBoolArrayValue(output_, offset_, value)) {
            offset_++;
            return true;
        }

        return false;
    }

    static bool EvmRevertDecoder(const bytes& reciept_output, std::string& ret) {
        bytes tmp = reciept_output;
        bytes magic ;
        magic.push_back(0x08);
        magic.push_back(0xc3);
        magic.push_back(0x79);
        magic.push_back(0xa0);
        if (reciept_output.size() > 4 &&
            bytes(reciept_output.begin(), reciept_output.begin() + 4) == magic) {
            tmp = bytes(reciept_output.begin() + 4, reciept_output.end());
        }

        EVMOutput out;
        out.SetOutput(tmp);
        if (!out.ReadString(ret)) {
            return false;
        }
        return true;
    }
};

}  // namespace mychain
#endif
