/**
 * Copyright 2018 Ant Financial Group. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0(the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Author : dongwei.ldw@antfin.com
 * Date : 2019-01-16
 */

#pragma once

#include <utility>
#include "serialize.hpp"
#include "tools.h"
#include "types.h"

namespace mychain {
class Identity {
public:
    Identity() : data_(kConstIdLength, 0) {}

    Identity(const char* str) : Identity(std::string(str)) {}

    Identity(const std::string& id) {
        if (id.size() > kConstIdLength) {
            data_ = Hex2Bin(id);
        } else {
            data_ = id;
        }
        MyAssert(data_.size() == kConstIdLength,
                 "MYCHAINLIB_ERROR: bin format Identity has wrong size");
    }

    Identity(const Identity& rl) { data_ = rl.data_; }

    Identity(Identity&& rl) { data_ = std::move(rl.data_); }

    Identity& operator=(const Identity& rl) {
        if (this != &rl) {
            data_ = rl.data_;
        }
        return *this;
    }

    Identity& operator=(Identity&& rl) {
        data_ = std::move(rl.data_);
        return *this;
    }

    inline const std::string& get_data() const { return data_; }

    inline std::string to_hex(bool upppercase = false) const {
        return Bin2Hex(data_, upppercase);
    }

private:
    std::string data_;

    static const uint32_t kConstIdLength = IdentityLength;

    // Serialize a Identity into a stream
    template <typename D>
    friend datastream<D>& operator<<(datastream<D>& ds, const Identity& v) {
        ds << v.data_;
        return ds;
    }

    // Deserialize a Identity from a stream
    template <typename D>
    friend datastream<D>& operator>>(datastream<D>& ds, Identity& v) {
        unsigned_int s;
        ds >> s;
        stream_assert(kConstIdLength == s.value,
                      ds.err_prefix() +
                          "deserialize using datastream,size of Identity's data field "
                          "and unpacked size don't match");
        ds.read(v.data_.data(), kConstIdLength);
        return ds;
    }

    friend bool operator<(const Identity& lhs, const Identity& rhs) {
        return lhs.data_ < rhs.data_;
    }
    friend bool operator==(const Identity& lhs, const Identity& rhs) {
        return lhs.data_ == rhs.data_;
    }
    friend bool operator!=(const Identity& lhs, const Identity& rhs) {
        return lhs.data_ != rhs.data_;
    }
};
};  // namespace mychain
